from threading import Thread
from queue import Queue
from socket import socket
from logging import getLogger


class Sender(Thread):
    def __init__(self, connection_queue: "Queue[Event]", connection: socket, connection_id: int) -> None:
        super(Sender, self).__init__()
        self.connection_queue = connection_queue
        self.connection = connection
        self.logger = getLogger(f"Sender - {connection_id}")

    def run(self) -> None:
        while True:
            message = self.connection_queue.get()
            self.logger.debug("sending %s",  len(message))
            message_size = len(message)
            try:
                self.connection.send(message_size.to_bytes(8, "big", signed=False) + message)
            except BrokenPipeError:
                self.connection_queue.put(message)
                return
            except Exception:
                self.connection_queue.put(message)
                return
            self.logger.debug("send %s",  len(message))



from logging import getLogger
from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection
from multiprocessing.queues import Queue as MPQueue
from pickle import dumps, loads
from queue import Queue
from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR, SOL_SOCKET, SO_KEEPALIVE, IPPROTO_TCP, TCP_KEEPIDLE, \
    TCP_KEEPINTVL, TCP_KEEPCNT
from threading import Thread
from time import sleep
from typing import Optional, Type, Dict, Set
from uuid import uuid4

from edri.dataclass.event import Event
from edri.config.constant import SWITCH_NEW_DEMANDS, SWITCH_LAST_MESSAGES
from edri.config.setting import SWITCH_HOST, SWITCH_PORT
from edri.dataclass.event import BaseIdentification
from edri.events.edri.router import SubscribeConnector, SubscribedAll, SubscribedNew, \
    LastEvents, Demands as RouterDemands, SendFrom
from edri.events.edri.switch import Demands as SwitchDemands, LastMessages as SwitchLastMessages
from edri.utility.connector import Socket


class Connector(Process):
    def __init__(self, router_queue: MPQueue) -> None:
        super(Connector, self).__init__()
        self.router_queue = router_queue
        self.local_request: Set[Type[Event]] = set()
        self.local_response: Set[Type[Event]] = set()
        self.remote_request: Dict[Type[Event], Set[int]] = {}
        self.remote_response: Dict[Type[Event], Set[int]] = {}
        self.logger = getLogger("SwitchConnector")
        self.router_pipe: Optional[Connection] = None

        self.switch_socket: Optional[socket] = None

        self.sender_queue: Optional[Queue] = None
        self.sender_thread: Optional[Thread] = None
        self.receiver_thread: Optional[Thread] = None

        self.switch_id = uuid4()
        self.socket = Socket()

    def sender(self) -> None:
        while True:
            message: Event = self.sender_queue.get()
            if message is False:
                break
            message._switch = BaseIdentification(id=self.switch_id)
            self.logger.debug("Sender - %s", message.__class__)
            if message.get_response() is None:
                router_id_list = self.remote_request.get(message.__class__, None)
            else:
                router_id_list = self.remote_response.get(message.__class__, None)
            for router_id in router_id_list:
                self.logger.debug("Odesílání zprávy switchi %s: %s", router_id, message)
                data = router_id.to_bytes(8, "big", signed=False)
                message_bytes = dumps(message)
                data += len(message_bytes).to_bytes(8, "big", signed=False)
                data += message_bytes
                try:
                    self.switch_socket.send(data)
                except BrokenPipeError:
                    self.logger.warning("Zpráva předána do fronty: %s", message)
                    self.sender_queue.put(message)
                    return
        self.logger.error("Sender se ukončil")

    def receiver(self) -> None:
        def load(amount: int) -> Optional[bytes]:
            data = bytes()
            while True:
                data += self.switch_socket.recv(amount - len(data))
                if not data:
                    return None
                elif len(data) == amount:
                    return data

        while True:
            message_size_bytes = load(8)
            if not message_size_bytes:
                self.switch_socket.shutdown(SHUT_RDWR)
                self.switch_socket.close()
                self.logger.error("Spojení se switchem ukončeno")
                return
            message_size = int.from_bytes(message_size_bytes, "big", signed=False)
            self.logger.debug("Příjem zprávy o velikost: %s", message_size)
            message_bytes = load(message_size)
            message: Event = loads(message_bytes)
            self.logger.debug("Příjem zprávy: %s", message)
            if isinstance(message, SwitchDemands):
                self.logger.debug("Přijaty požadavky od switche: %s", message)
                for demand in message.demands.requests:
                    if demand in self.remote_request:
                        self.remote_request[demand].add(message.connection_id)  # je to set, takže na klid
                    else:
                        self.remote_request[demand] = {message.connection_id}
                        subscribe = SubscribeConnector(event=demand, request=True)
                        self.router_queue.put(subscribe)

                for demand in message.demands.responses:
                    if demand in self.remote_response:
                        self.remote_response[demand].add(message.connection_id)  # je to set, takže na klid
                    else:
                        self.remote_response[demand] = {message.connection_id}
                        subscribe = SubscribeConnector(event=demand, request=False)
                        self.router_queue.put(subscribe)
            elif isinstance(message, SwitchLastMessages):
                last_message = message.last_messages.get(self.switch_id, None)
                self.router_queue.put(SendFrom(id=last_message))
            else:
                message._switch = True
                self.logger.info("-> Router %s", message)
                self.router_queue.put(message)

    def send_demands(self) -> None:
        self.logger.info("Odesílání požadavků do switche")
        data = SWITCH_NEW_DEMANDS
        local_messages_bytes = dumps(RouterDemands(
            requests=self.local_request,
            responses=self.local_response
        ))
        data += len(local_messages_bytes).to_bytes(8, "big", signed=False)
        data += local_messages_bytes
        self.switch_socket.send(data)

    def request_last_messages(self) -> None:
        self.router_queue.put(LastEvents())

    def send_last_messages(self, message: LastEvents) -> None:
        self.logger.info("Odesílání žádosti o poslední zprávy")
        data = SWITCH_LAST_MESSAGES
        last_messages_bytes = dumps(message)
        data += len(last_messages_bytes).to_bytes(8, "big", signed=False)
        data += last_messages_bytes
        self.switch_socket.send(data)

    def receive_setting(self) -> None:
        self.logger.debug("Sending message to receive basic setting")
        self.switch_socket.send()

    def send_basic_info(self) -> None:
        self.logger.debug("Odesílání základních dat switchi")
        data = self.switch_id.to_bytes(8, "big", signed=False)
        self.switch_socket.send(data)
        self.send_demands()
        self.request_last_messages()

    def connect(self) -> None:
        while True:
            self.logger.debug("Připojování ke switchi")
            self.switch_socket = socket(AF_INET, SOCK_STREAM)
            self.switch_socket.settimeout(10)
            self.switch_socket.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
            try:
                self.switch_socket.connect((SWITCH_HOST, SWITCH_PORT))
                self.switch_socket.settimeout(None)
                self.logger.info("Připojeno ke switchi")
                break
            except ConnectionRefusedError as e:
                self.logger.warning("Nelze se připojit ke switchi - připojení odmítnuto", exc_info=e)
                sleep(5)
            except TimeoutError as e:
                self.logger.warning("Nelze se připojit ke switchi - připojování trvalo příliš dlouho", exc_info=e)
                sleep(5)
            except Exception as e:
                self.logger.warning("Nelze se připojit ke switchi - neznámá chyba", exc_info=e)
                sleep(10)

    def run(self) -> None:
        sleep(10)
        self.logger.debug(f"SwitchConnector se spustil!")
        self.connect()

        self.router_pipe, pipe = Pipe(duplex=False)
        subscribed = SubscribedAll(pipe=pipe)
        self.router_queue.put(subscribed)
        while True:
            message = self.router_pipe.recv()
            if isinstance(message, SubscribedAll):
                self.local_request = message.get_response().data["messages"].requests
                self.local_response = message.get_response().data["messages"].responses
                self.logger.debug("Zprávy routeru %s", message)
                break
            else:
                self.logger.error("Neočekávaná příchozí zpráva %s", message)
        pipe.close()

        self.send_basic_info()

        self.receiver_thread = Thread(target=self.receiver)
        self.receiver_thread.start()

        self.sender_queue = Queue()
        self.sender_thread = Thread(target=self.sender)
        self.sender_thread.start()

        while True:
            if self.router_pipe.poll(timeout=5):
                message = self.router_pipe.recv()
                self.logger.debug("<- Router - %s", message)
                if isinstance(message, SubscribedNew):
                    if message.request:
                        self.local_request.add(message.event)
                    else:
                        self.local_response.add(message.event)
                    self.send_demands()
                elif isinstance(message, LastEvents):
                    self.send_demands()
                    self.send_last_messages(message)
                else:
                    self.sender_queue.put(message)
            if not self.receiver_thread.is_alive():
                self.logger.error("Receiver se ukončil, navazuji nové spojení")
                self.sender_queue.put(False)
                sleep(5)
                self.connect()
                self.send_basic_info()
                self.receiver_thread = Thread(target=self.receiver)
                self.receiver_thread.start()

                self.sender_thread = Thread(target=self.sender)
                self.sender_thread.start()



from dataclasses import dataclass, field
from enum import IntEnum
from typing import Optional, TypeVar, Mapping, Any

T = TypeVar("T")


class ResponseStatus(IntEnum):
    NONE = -1
    OK = 0
    FAILED = 1


@dataclass
class Response:
    _status: ResponseStatus = field(init=False, default=ResponseStatus.NONE)
    _error: Optional[str] = field(init=False, default=None)
    _changed: bool | None = field(init=False, default=None)

    @property
    def dict(self) -> Mapping[str, str | ResponseStatus]:
        response = {"_response": self._status}
        if self._error:
            response["error"] = self._error

        return response

    def set_status(self, status: ResponseStatus) -> None:
        self._status = status

    def get_status(self) -> ResponseStatus:
        return self._status

    def __setattr__(self, key: str, value: Any) -> None:
        if self._changed is not None and not self._changed:
            super().__setattr__("_changed", True)
            super().__setattr__("_status", ResponseStatus.OK)
        super().__setattr__(key, value)

    def __post_init__(self) -> None:
        self._changed = False

    def has_changed(self) -> bool:
        return self._changed



from edri.dataclass.event import event
from edri.events.edri.group import Switch
from edri.events.edri.router import Demands as RouterDemands


@event
class Demands(Switch):
    connection_id: int
    demands: RouterDemands



from asyncio import Event, get_event_loop, create_task, wait, AbstractEventLoop
from json import dumps, loads
from multiprocessing import Process, Queue
from typing import Optional

from websockets import serve, WebSocketServerProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from dataclasses import asdict

from edri.abstract.api_base import APIBase
from edri.config.constant import ApiType
from edri.config.setting import HOST, WS_PORT
from edri.dataclass.event import api_events
from edri.dataclass.event.response import ResponseStatus
from edri.utility.function import snake2camel


class API(APIBase, Process):

    def __init__(self, ab_queue: "Queue[Event]") -> None:
        super().__init__(ab_queue)
        self.event_loop: Optional[AbstractEventLoop] = None
        self.events = {event.resource: event.event for event in api_events if ApiType.WS not in event.exclude}
        self.commands = {event.event: event.resource for event in api_events if ApiType.WS not in event.exclude}

    @property
    def name(self) -> str:
        return "Websocket API"

    @property
    def type(self) -> ApiType:
        return ApiType.WS

    async def new_client(self, websocket: WebSocketServerProtocol, url) -> None:
        self.logger.debug("New connection: %s %s", websocket.remote_address[0], url)
        for middleware in self.middlewares:
            if not middleware.process_req(websocket, url):
                return
        if (register := self.register()) is None:
            return
        client_pipe, key = register
        frame_available = Event()
        self.event_loop.add_reader(client_pipe.fileno(), frame_available.set)
        from_core = create_task(frame_available.wait())
        from_client = create_task(websocket.recv())
        tasks = {from_core, from_client}
        while True:
            try:
                done, pending = await wait(tasks, return_when="FIRST_COMPLETED")
                tasks = set()
                for task in done:
                    if task == from_core:
                        event = client_pipe.recv()
                        self.logger.debug(f"<- API Broker %s %s", key, event)
                        data = {
                            "command": self.commands[event.__class__]
                        }
                        data.update(asdict(event))
                        data.pop("_key")
                        data.pop("_switch")
                        data.pop("_stream")
                        data.pop("method", None)
                        if data["response"]:
                            status = data["response"].pop("_status")
                            data["response"]["event_status"] = status.value
                            data["response"].pop("_error")
                            data["response"].pop("_changed")
                            data["response"] = {snake2camel(key): value for key, value in data["response"].items() if not key.startswith("_")}
                        else:
                            data.pop("response")
                        await self.event_loop.create_task(websocket.send(dumps(data)))

                        from_core.cancel()
                        frame_available = Event()
                        self.event_loop.add_reader(client_pipe.fileno(), frame_available.set)
                        from_core = create_task(frame_available.wait())
                        tasks.add(from_core)
                    elif task == from_client:
                        ws_event: str = task.result()
                        self.logger.debug(f"-> %s %s", key, ws_event)
                        try:
                            ws_event = loads(ws_event)
                        except Exception as e:
                            await self.event_loop.create_task(
                                websocket.send(dumps(["Warning", {"warning": "Malformed JSON"}])))
                            self.logger.warning("Malformed JSON: %s", ws_event, exc_info=e)
                        else:
                            if not isinstance(ws_event, dict) or "command" not in ws_event:
                                await self.event_loop.create_task(
                                    websocket.send(dumps(ws_event)))
                                self.logger.warning("Unknown data: %s", ws_event)
                            else:
                                command = ws_event.pop("command")
                                if (event_type := self.events.get(command, None)) is None:
                                    ws_event["command"] = command
                                    ws_event["response"] = {
                                        "status": ResponseStatus.FAILED.value,
                                        "reason": "Unknown command"
                                    }
                                    await self.event_loop.create_task(
                                        websocket.send(dumps(ws_event)))
                                    self.logger.warning("Unknown command: %s", command)
                                else:
                                    try:
                                        event = event_type(**ws_event)
                                    except Exception as e:
                                        self.logger.warning("Cannot instantiated event: %s", command, exc_info=e)
                                        return
                                    event._key = key
                                    self.ab_queue.put(event)

                        finally:
                            from_client.cancel()
                            from_client = create_task(websocket.recv())
                            tasks.add(from_client)
                tasks = tasks | pending
                if len(tasks) == 0:
                    break

            except ConnectionClosedError as e:
                from_core.cancel()
                from_client.cancel()
                self.logger.info("ConnectionClosedError: %s", key, exc_info=e)
                break

            except ConnectionClosedOK as e:
                from_core.cancel()
                from_client.cancel()
                self.logger.debug("ConnectionClosedOK: %s", key, exc_info=e)
                break

            except ValueError as e:
                from_core.cancel()
                from_client.cancel()
                self.logger.warning("ValueError: %s", key, exc_info=e)
                break

        self.event_loop.remove_reader(client_pipe.fileno())
        self.unregister(client_pipe, key)
        return

    def run(self) -> None:
        self.event_loop = get_event_loop()
        start_server = serve(self.new_client, HOST, WS_PORT)

        self.event_loop.run_until_complete(start_server)
        self.logger.info("WebSocket API has started!")
        try:
            self.event_loop.run_forever()
        except KeyboardInterrupt:
            return



from edri.dataclass.event import event, Response, response
from edri.events.edri.group import Router


@response
class HealthCheckResponse(Response):
    name: str


@event
class HealthCheck(Router):
    response: HealthCheckResponse




from typing import Type, Set

from edri.dataclass.event import Event, event
from edri.events.edri.group import Router


@event
class Demands(Router):
    requests: Set[Type[Event]]
    responses: Set[Type[Event]]



from abc import ABC
from multiprocessing import Process
from multiprocessing.connection import Connection
from typing import TypeVar

from edri.dataclass.event import Event
from edri.abstract.worker import Worker

T = TypeVar('T', bound=Event)


class WorkerProcess(Worker[T], Process, ABC):
    """
    A WorkerProcess class that integrates functionality from a custom Worker
    class with multiprocessing capabilities, designed to run as a separate
    process. This class is abstract and should be subclassed to implement
    specific worker behavior in a multiprocessing environment.

    Attributes:
        Inherits all attributes from the Worker class and the Process class.

    Methods:
        __init__(pipe: Connection, event, name: str): Initializes a new instance
        of the WorkerProcess.

    Parameters:
        pipe (Connection): A connection object used for communication between
        processes. This should be an instance of multiprocessing.Connection.
        event: The event type that the worker will handle. The specific type is
        determined by the EventType generic type variable, which should be
        defined in the subclass.
        name (str): A name for the process. This is used internally by the
        multiprocessing module to identify the process, especially useful for
        debugging.

    Returns:
        None: The constructor initializes the process but does not return a value.

    Raises:
        Inherits any exceptions raised by the Worker class or the Process class
        constructors.

    Note:
        As this class is abstract, it cannot be instantiated directly. It requires
        subclassing and implementation of the abstract methods defined in the
        Worker class.
    """

    def __init__(self, pipe: Connection, event, name: str) -> None:
        Worker.__init__(self, pipe, event, name)
        Process.__init__(self)



from abc import abstractmethod
from time import sleep
from logging import Logger
from multiprocessing import Queue
from multiprocessing.connection import wait, Connection, Pipe
from queue import PriorityQueue
from threading import Thread
from typing import Type, Optional, Callable

from edri.abstract import ManagerBase
from edri.dataclass.event import Event
from edri.dataclass.queue_item import QueueItem
from edri.dataclass.worker import Worker
from edri.events.edri.group import Manager
from edri.events.edri.manager import StreamCreate, StreamMessage, StreamClose, WorkerQuit


class ManagerPriorityBase(ManagerBase):
    """
    An extension of ManagerBase that introduces priority handling for events. This class
    manages events using a priority queue, ensuring that higher-priority events are processed
    before lower-priority ones.

    Attributes:
        event_queue (PriorityQueue[QueueItem]): A priority queue for managing events based on their priority.
        event_worker (Thread): A thread dedicated to handling command execution based on event priority.
        waking_up (tuple[Connection, Connection]): A pair of connections used to wake up the manager when a new worker starts.

    Methods:
        __init__: Initializes the ManagerPriorityBase with an optional router queue and logger.
        _prepare: Prepares the manager by initializing the event queue, event worker thread, and waking up pipe.
        additional_pipes: Property to include the waking_up pipe in the set of additional pipes the manager listens to.
        command_handler: Thread target function that processes events from the priority queue.
        get_priority: Abstract method that must be implemented to define the priority of events.
        run_resolver: Main event loop, enhanced to handle events based on priority.
        start_worker: Overrides the start_worker method to wake up the manager when a new worker is started.
    """

    def __init__(self, router_queue: Optional["Queue[Event]"] = None, logger: Optional[Logger] = None) -> None:
        """
        Initializes the ManagerPriorityBase with an optional router queue and logger, setting up the foundation
        for event priority management.

        Parameters:
            router_queue (Optional[Queue[Event]]): The queue for communicating with the router.
            logger (Optional[Logger]): Logger instance for logging messages.
        """
        super().__init__(router_queue, logger)
        self.event_queue: "PriorityQueue[QueueItem]" = None  # type: ignore
        self.event_worker: Thread = None  # type: ignore
        self.waking_up: tuple[Connection, Connection] = None  # type: ignore

    def _prepare(self) -> None:
        """
        Prepares the manager by initializing the event queue, starting the event worker thread,
        and setting up the waking_up pipe used to signal new work.
        """
        super()._prepare()
        self.event_queue = PriorityQueue()
        self.event_worker = Thread(target=self.command_handler)
        self.event_worker.start()
        self.waking_up = Pipe(duplex=False)

    @property
    def additional_pipes(self) -> set[Connection]:
        """
        Includes the waking_up pipe in the set of additional pipes the manager listens to.

        Returns:
            set[Connection]: A set of additional pipes, including the waking_up pipe.
        """
        return {self.waking_up[0]}

    def command_handler(self) -> None:
        """
         The thread target function that processes events from the priority queue, ensuring that events are handled
         in order of their priority.
         """
        while True:
            event = self.event_queue.get().item
            self.resolve(event)

    @abstractmethod
    def get_priority(self, event: Type[Event]) -> int:
        """
        Abstract method that determines the priority of an event. Must be implemented by subclasses
        to define how event priorities are assigned.

        Parameters:
            event (Type[Event]): The class of the event for which the priority needs to be determined.

        Returns:
            int: The priority of the event.
        """
        pass

    def run_resolver(self) -> None:
        """
        The main event loop for the manager, enhanced to handle events based on priority. It listens on all pipes,
        including the waking_up pipe, to efficiently process and prioritize events.
        """
        while True:
            pipes = self.get_pipes()
            if not pipes:
                sleep(1)
                continue
            try:
                active_pipes: list[Connection] = wait(pipes, timeout=10)
                for active_pipe in active_pipes:
                    if active_pipe == self.router_pipe:
                        event: Event = self.router_pipe.recv()
                        self.logger.debug("Přijata zpráva z routeru: %s", event)
                        priority = self.get_priority(event.__class__)
                        self.event_queue.put(QueueItem(priority, event))
                    elif active_pipe == self.waking_up[0]:
                        self.waking_up[0].recv()
                        self.logger.debug("Načtení nových rour")
                        continue
                    else:
                        try:
                            event = active_pipe.recv()
                            self.logger.debug("Přijata zpráva: %s", event)
                        except EOFError as e:
                            self.logger.error("Problém s komunikací", exc_info=e)
                            continue
                        except OSError as e:
                            self.logger.error("Problém v OS", exc_info=e)
                            continue
                        try:
                            key, worker = self._find_worker(active_pipe)
                            if isinstance(event, Manager):
                                if isinstance(event, StreamCreate):
                                    self._resolve_callback_stream_create(event, worker)
                                elif isinstance(event, StreamMessage):
                                    self._resolve_callback_stream_message(event, worker)
                                elif isinstance(event, StreamClose):
                                    self._resolve_callback_stream_close(event, worker)
                                elif isinstance(event, WorkerQuit):
                                    self._remove_worker(key)
                            else:
                                self.resolve_callback_worker(event, worker)
                        except ValueError:
                            self.resolve_callback_pipe(event, active_pipe)
            except KeyboardInterrupt:
                return

    def start_worker(self, event: Event, resolver: Callable, *args, **kwargs) -> Worker:
        """
        Starts a worker to handle an event and signals the manager that new work has started,
        potentially affecting the priority queue.

        Parameters:
            event (Event): The event to be handled by the worker.
            resolver (Callable): The function that will handle the event.
            *args: Additional positional arguments for the resolver.
            **kwargs: Additional keyword arguments for the resolver.

        Returns:
            Worker: The worker instance that was started to handle the event.
        """
        worker = super().start_worker(event, resolver, *args, **kwargs)
        self.waking_up[1].send(None)
        return worker



from logging import getLogger
from multiprocessing import Queue
from typing import Dict, Any

from edri.dataclass.event import Event
from edri.abstract import ManagerBase
from edri.events.edri.store import Set, Get


class Store(ManagerBase):
    def __init__(self, router_queue: Queue[Event]) -> None:
        super().__init__(router_queue, getLogger(__name__))
        self.store: Dict[str, Any] = {}

    def solve_req_set(self, event: Set) -> None:
        self.store[event.name] = event.value

    def solve_req_get(self, event: Get) -> None:
        event.response.value = self.store.get(event.name, None)



import asyncio
from dataclasses import asdict
from enum import Enum
from json import JSONEncoder, dumps, loads
from logging import getLogger
from multiprocessing import Process
from multiprocessing.queues import Queue
from os import remove
from pathlib import Path
from tempfile import NamedTemporaryFile
from types import NoneType
from typing import Type, Any, Union, Pattern
from urllib.parse import parse_qs, quote
from re import compile
from http.cookies import SimpleCookie

from uvicorn import Config, Server

from edri.config.setting import UPLOAD_FILES_PREFIX, REST_RESPONSE_TIMEOUT, REST_PORT
from edri.abstract.api_base import APIBase
from edri.config.constant import ApiType
from edri.dataclass.event import Method, Event, api_events
from edri.dataclass.event.response import ResponseStatus
from edri.utility.function import snake2camel, camel2snake


class REST(Process):
    """
    A process dedicated to running a RESTful API server using Uvicorn, an ASGI server.
    Inherits from multiprocessing.Process to run the API server in a separate process.

    Attributes:
        api_broker_queue (Queue): A multiprocessing queue used for communication with
        the API broker, facilitating the handling of API requests.
        type (ApiType): The type of the API, set to ApiType.REST for RESTful APIs.
        logger (Logger): Logger instance for logging information related to API server
        operations.

    Methods:
        run(): Starts the Uvicorn server with the provided configuration and runs
        the ASGI application to handle RESTful API requests.
    """

    def __init__(self, api_broker_queue: Queue[Event]) -> None:
        """
        Initializes the REST API server process.

        Parameters:
            api_broker_queue (Queue): The queue for communication with the API broker.
        """
        super().__init__()
        self.api_broker_queue = api_broker_queue

        self.type = ApiType.REST
        self.logger = getLogger(__name__)

    class App(APIBase):
        """
        An ASGI application for handling RESTful API requests, translating them
        into events for processing by the system.

        Inherits from APIBase to utilize common API functionalities and to handle
        requests in a structured manner.

        Attributes:
            events (dict): A dictionary mapping API paths to events and their handling
            methods.
            http_codes_mapping (dict): Maps response statuses to HTTP status codes.
        """

        def __init__(self, ab_queue: Queue[Event]) -> None:
            """
            Initializes the ASGI application with a queue for communication and
            predefines the mapping of API events.

            Parameters:
                ab_queue (Queue): The queue for communication with the API broker.
            """
            super().__init__(ab_queue)
            self.events = self.sort_events()
            self.http_codes_mapping = {
                ResponseStatus.OK: 200,
                ResponseStatus.FAILED: 409,
            }

        @staticmethod
        def sort_events() -> dict[str, dict[Method | str, Type[Event] | Pattern[str]]]:
            """
            Organizes API events by their URL and HTTP method for efficient lookup.

            Returns:
                A dictionary organizing events by URL and method.
            """
            data: dict[str, dict] = {}
            url_replace = compile(r"{([^}]+)}")
            for api_event in api_events:
                position = api_event.url.find("{")
                if position == -1:
                    url = api_event.url
                else:
                    url = api_event.url[:position]
                if url not in data:
                    data[url] = dict()
                events = data[url]
                if ApiType.REST in api_event.exclude:
                    continue
                for method in api_event.event.method:
                    if method in events:
                        raise ValueError("Method %s was already set.", method)
                    events[method] = api_event.event
                    if position != -1:
                        events["url"] = compile(url_replace.sub(r"(?P<\1>.+)", api_event.url))
            return data

        @staticmethod
        def get_default_headers(headers: list[tuple[bytes, bytes]], content_type: bytes = b"application/json") -> list[tuple[bytes, bytes]]:
            response_headers: list[tuple[bytes, bytes]] = [
                (b"Access-Control-Allow-Headers", b"Accept,Accept-Language,Content-Language,Content-Type,Filename"),
                (b"Content-Type", content_type)
            ]
            for name, value in headers:
                if name == b"origin":
                    response_headers.append((b"Access-Control-Allow-Origin", value))
                    break

            return response_headers

        @property
        def name(self) -> str:
            """
            Returns the name of the REST API component, used primarily for logging purposes.

            Returns:
                The name of the API component.
            """
            return "edri.api.rest"

        @property
        def type(self) -> ApiType:
            """
            Identifies the API type as REST, aligning with the application's architecture.

            Returns:
                The enumerated type indicating this component handles RESTful API requests.
            """
            return ApiType.REST

        @staticmethod
        def dict_factory(data: list[tuple[str, Any]]) -> dict:
            """
            Converts a list of tuples into a dictionary, handling specific types like Enums
            by converting them to their value.

            Parameters:
                data (list[tuple[str, Any]]): A list of key-value pairs.

            Returns:
                A dictionary with keys mapped to converted values if necessary.
            """

            def convert_value(obj: Any) -> Any:
                if isinstance(obj, Enum):
                    return obj.value
                return obj

            return {key: convert_value(value) for key, value in data}

        class CustomJSONEncoder(JSONEncoder):
            """
            A custom JSON encoder for converting various data types to JSON-compatible
            formats, including support for Enums, datetime objects, Paths, bytes, and
            more.

            Inherits from JSONEncoder to override the default() method for custom serialization.
            """

            def default(self, data) -> Any:
                if hasattr(data, "to_json"):
                    return data.to_json()
                elif hasattr(data, "isoformat"):
                    return data.isoformat()
                elif isinstance(data, Path):
                    return data.as_posix()
                elif isinstance(data, bytes) or isinstance(data, bytearray):
                    return data.hex()
                elif isinstance(data, Enum):
                    return data.value
                else:
                    return super().default(data)

        def check_type(self, values: Any, annotation: Any) -> Any:
            """
            Validates and converts input values to the specified annotation type,
            supporting basic types, Optional, and lists with type annotations.

            Parameters:
                values: The input value to be validated and converted.
                annotation: The target type annotation for the conversion.

            Returns:
                The converted value if conversion is successful.

            Raises:
                TypeError: If the value cannot be converted to the specified type.
            """
            if hasattr(annotation, "__origin__"):
                if annotation.__origin__ == Union:
                    try:
                        none_position = annotation.__args__.index(NoneType)
                    except ValueError:
                        raise TypeError("Unions can consist only of basic type and None")
                    if values is None:
                        return values
                    else:
                        return self.check_type(values, annotation.__args__[0 if none_position == 1 else 1])
                elif annotation.__origin__ == list and isinstance(values, list):
                    if len(annotation.__args__) != 1:
                        raise TypeError("Type of list item must be specified")
                    return [self.check_type(value, annotation.__args__[0]) for value in values]
                else:
                    raise TypeError("Value '%s' cannot be converted to type %s" % (values, annotation))
            else:
                try:
                    return annotation(values)
                except Exception as e:
                    if hasattr(annotation, "fromisoformat"):
                        try:
                            return annotation.fromisoformat(values)
                        except Exception as e:
                            self.logger.debug("Convert error", exc_info=e)
                            raise TypeError("Value '%s' cannot be converted from isoformat to type %s" % (values, annotation))
                    else:
                        self.logger.debug("Convert error", exc_info=e)
                        raise TypeError("Value '%s' cannot be converted to type %s" % (values, annotation))

        async def __call__(self, scope, receive, send) -> None:
            """
            The main entry point for handling incoming HTTP requests as per the ASGI
            specification. Routes requests to the appropriate event based on the URL
            and method, performs data validation, and sends HTTP responses.

            Parameters:
                scope: The scope of the incoming request containing type, path, headers, etc.
                receive: An awaitable callable yielding incoming messages.
                send: An awaitable callable taking outgoing messages.

            Returns:
                None: This method handles the request and sends an HTTP response
                without returning a value.
            """
            if scope["type"] == "lifespan":  # waiting for startup
                message = await receive()
                if message["type"] == "lifespan.startup":
                    await send({"type": "lifespan.startup.complete"})
                elif message["type"] == "lifespan.shutdown":
                    await send({"type": "lifespan.shutdown.complete"})

            elif scope["type"] == "http":
                loop = asyncio.get_event_loop()
                event_constructors = self.events.get(scope["path"], None)
                parameters: dict[str, str] = {}
                if not event_constructors:
                    for url, events in self.events.items():
                        url_re = events.get("url", None)
                        if not url_re:
                            continue
                        if scope["path"].startswith(url):
                            match = url_re.match(scope["path"])
                            parameters.update(match.groupdict())
                            event_constructors = events
                            break
                    else:
                        self.logger.debug("Unknown url %s", scope["path"])
                        await send({
                            'type': 'http.response.start',
                            'status': 400,
                            'headers': [
                                [b'content-type', b'text/plain'],
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b"",
                            'more_body': False
                        })
                        return

                if scope["method"] == "OPTIONS":
                    # to vecpané URL :|
                    headers = self.get_default_headers(scope["headers"])
                    headers.append(
                        (b"Access-Control-Allow-Methods",
                         ", ".join(x.value for x in event_constructors.keys() if x != "url").encode())
                    )
                    await send({
                        'type': 'http.response.start',
                        'status': 200,
                        'headers': headers
                    })
                    await send({
                        'type': 'http.response.body',
                        'body': b"",
                        'more_body': False
                    })
                    return

                method = Method(scope["method"])
                event_constructor = event_constructors.get(method, None)

                if not event_constructor:
                    await send({
                        'type': 'http.response.start',
                        'status': 405,
                        'headers': [
                            [b'content-type', b'text/plain'],
                        ]
                    })
                    await send({
                        'type': 'http.response.body',
                        'body': b"",
                        'more_body': False
                    })
                    return

                cookies = None
                for key, value in scope["headers"]:
                    if key == b"cookie":
                        simple_cookies = SimpleCookie()
                        simple_cookies.load(value.decode())
                        cookies = {key: value.value for key, value in simple_cookies.items()}
                        break
                url_parameters = parse_qs(scope["query_string"].decode("utf-8"), keep_blank_values=True)
                parameters.update(
                    {camel2snake(key.strip("[]")): value if key.endswith("[]") else value[-1] for key, value in
                     url_parameters.items()})

                json_request = False
                for key, value in scope["headers"]:
                    if key == b"content-type" and value == b"application/json":
                        json_request = True

                if json_request:
                    raw_data = b""
                    while True:
                        if len(raw_data) > 4096 * 1024:
                            await send({
                                'type': 'http.response.start',
                                'status': 413,
                                'headers': [
                                    [b'content-type', b'text/plain'],
                                ]
                            })
                            await send({
                                'type': 'http.response.body',
                                'body': b"",
                                'more_body': False
                            })
                            return

                        body = await receive()
                        data = body.get("body", None)
                        if data is not None:
                            raw_data += data
                            if not body["more_body"]:
                                break
                        else:
                            break

                    if not raw_data:  # no data = error in json parser - prepare dummy data for it
                        raw_data = b"{}"

                    try:
                        parameters.update({camel2snake(key): value for key, value in loads(raw_data).items()})
                    except Exception as e:
                        self.logger.warning("Cannot process json data", exc_info=e)
                        await send({
                            'type': 'http.response.start',
                            'status': 422,
                            'headers': [
                                [b'content-type', b'text/plain'],
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b"",
                            'more_body': False
                        })
                        return

                else:
                    for key, value in scope["headers"]:
                        if key == b"content-length":
                            try:
                                size = int(value)
                            except Exception:
                                continue
                            if size > 0:
                                temporary_file = NamedTemporaryFile(prefix=UPLOAD_FILES_PREFIX, delete=False)
                                parameters["path"] = Path(temporary_file.name)
                                break

                for name, value in parameters.items():
                    annotation = event_constructor.__annotations__.get(name, None)
                    if not annotation:
                        self.logger.debug("Unknown parameter %s", name)
                        await send({
                            'type': 'http.response.start',
                            'status': 400,
                            'headers': [
                                [b'content-type', b'text/plain'],
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b"",
                            'more_body': False
                        })
                        return
                    try:
                        value = self.check_type(value, annotation)
                    except TypeError as e:
                        self.logger.debug("Wrong type %s for %s:%s", type(value), name, annotation, exc_info=e)
                        await send({
                            'type': 'http.response.start',
                            'status': 400,
                            'headers': [
                                [b'content-type', b'text/plain'],
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b"",
                            'more_body': False
                        })
                        return
                    except Exception as e:
                        self.logger.debug("Unknown error during type checking", exc_info=e)
                        await send({
                            'type': 'http.response.start',
                            'status': 500,
                            'headers': [
                                [b'content-type', b'text/plain'],
                            ]
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b"",
                            'more_body': False
                        })
                        return
                    parameters[name] = value

                try:
                    event = event_constructor(**parameters)
                except Exception as e:
                    self.logger.warning("Cannot create event", exc_info=e)
                    if not json_request and "path" in parameters:
                        temporary_file.close()
                        remove(temporary_file.name)
                    await send({
                        'type': 'http.response.start',
                        'status': 400,
                        'headers': [
                            [b'content-type', b'text/plain'],
                        ]
                    })
                    await send({
                        'type': 'http.response.body',
                        'body': b"",
                        'more_body': False
                    })
                    return

                if not json_request and "path" in parameters:
                    while True:
                        body = await receive()
                        data = body.get("body", None)
                        if data is not None:
                            temporary_file.write(data)
                            if not body["more_body"]:
                                break
                        else:
                            break
                    temporary_file.close()

                event._headers = scope["headers"]
                event._cookies = cookies
                if (register := self.register()) is None:
                    self.logger.error("Registration failed")
                    await send({
                        'type': 'http.response.start',
                        'status': 500,
                        'headers': self.get_default_headers(scope["headers"]),

                    })
                    return
                pipe, key = register
                event._key = key
                await loop.run_in_executor(None, self.ab_queue.put, event)
                available_data = await loop.run_in_executor(None, pipe.poll, REST_RESPONSE_TIMEOUT)
                if not available_data:
                    self.logger.error("No response was returned in %ss", REST_RESPONSE_TIMEOUT)
                    await send({
                        'type': 'http.response.start',
                        'status': 500,
                        'headers': self.get_default_headers(scope["headers"]),
                        'more_body': False

                    })
                    return

                event_response = pipe.recv()

                response = event_response.get_response()

                path = getattr(response, "path", None)
                if isinstance(path, Path):
                    headers = self.get_default_headers(scope["headers"], b"application/octet-stream")
                    headers.append((b"Content-Disposition", b'attachment; filename="' + response.name.encode("ascii",
                                                                                                             errors="ignore") + b'"; filename*=UTF-8\'\'' + quote(
                        response.name, safe="").encode()))
                    headers.append((b"Content-Length", str(path.lstat().st_size).encode()))
                    headers.append((b"Content-Type", response.content_type.encode()))
                    try:
                        with path.open("rb") as file:
                            await send({
                                'type': 'http.response.start',
                                'status': self.http_codes_mapping.get(response.get_status(), 500),
                                'headers': headers,
                            })
                            while True:
                                data = file.read(1024)
                                if not data:
                                    break
                                await send({
                                    'type': 'http.response.body',
                                    'body': data,
                                    'more_body': True
                                })

                            await send({
                                'type': 'http.response.body',
                                'body': b'',
                                'more_body': False
                            })
                    except FileNotFoundError:
                        self.logger.warning("File to download was not found: %s", path)
                        await send({
                            'type': 'http.response.start',
                            'status': 500,
                            'headers': headers,
                        })
                        await send({
                            'type': 'http.response.body',
                            'body': b'',
                            'more_body': False
                        })
                        return
                else:
                    await send({
                        'type': 'http.response.start',
                        'status': self.http_codes_mapping.get(response.get_status(), 500),
                        'headers': self.get_default_headers(scope["headers"]),

                    })

                    body = asdict(response, dict_factory=self.dict_factory)
                    body.update(response.dict)
                    body = {snake2camel(key): value for key, value in body.items() if not key.startswith("_")}

                    await send({
                        'type': 'http.response.body',
                        'body': dumps(body, ensure_ascii=False, cls=self.CustomJSONEncoder).encode("utf-8"),
                        'more_body': False
                    })

                self.unregister(pipe, key)

    def run(self) -> None:
        """
        Configures and starts the Uvicorn server to run the ASGI application for
        handling RESTful API requests.
        """
        config = Config(self.App(self.api_broker_queue), host="0.0.0.0", port=REST_PORT, log_level="debug", workers=8)
        # ssl_keyfile="/home/marek/rest.key",
        # ssl_certfile="/home/marek/rest.crt")
        server = Server(config)
        server.run()



from logging import getLogger
from multiprocessing import Queue
from typing import Dict, Any

from edri.dataclass.event import Event
from edri.abstract import ManagerBase
from edri.events.edri.store import Set, Get, Delete


class Store(ManagerBase):
    """
    A simple key-value store that processes set and get requests for storing and retrieving
    data, respectively. Inherits from ManagerBase to integrate with the system's event-driven
    architecture.

    Attributes:
        router_queue (Queue): The messaging queue for receiving and sending events.
        store (Dict[str, Any]): A dictionary acting as the internal store for key-value pairs.
        logger (Logger): Logger instance for logging operations within the Store.

    Methods:
        solve_req_set(event: Set): Handles requests to set (store) a value associated with a key.
        solve_req_get(event: Get): Handles requests to get (retrieve) a value by its key.
    """
    def __init__(self, router_queue: "Queue[Event]") -> None:
        """
        Initializes the Store with a router queue for communication and an empty dictionary
        for storing key-value pairs.

        Parameters:
            router_queue (Queue): The messaging queue for communicating with other system components.
        """
        super().__init__(router_queue, getLogger(__name__))
        self.store: Dict[str, Any] = {}

    def solve_req_set(self, event: Set) -> None:
        """
        Processes a set request by storing the provided value associated with the specified key
        in the internal store.

        Parameters:
            event (Set): The event containing the key (name) and the value to be stored.
        """
        self.store[event.name] = event.value
        self.logger.debug("Value set for key '%s': %s", event.name, event.value)

    def solve_req_get(self, event: Get) -> None:
        """
        Processes a get request by retrieving the value associated with the specified key from
        the internal store. If the key does not exist, None is returned.

        Parameters:
            event (Get): The event containing the key (name) for which the value is requested.
        """
        event.response.value = self.store.get(event.name, None)
        if event.response.value is not None:
            self.logger.debug("Value retrieved for key '%s': %s", event.name, event.response.value)
        else:
            self.logger.debug("No value found for key '%s'", event.name)

    def solve_req_delete(self, event: Delete) -> None:
        """
        Processes a delete request by removing the specified key and its associated value
        from the internal store. If the key does not exist, no action is taken.

        Parameters:
            event (Delete): The event containing the key (name) of the item to be deleted.
        """
        if event.name in self.store:
            del self.store[event.name]
            self.logger.debug("Deleted item with key '%s'", event.name)
        else:
            self.logger.debug("Attempted to delete non-existent key '%s'", event.name)



from logging import getLogger
from pickle import dumps
from threading import Thread
from typing import Dict
from queue import Queue


from edri.dataclass.switch import Connection
from edri.events.edri.switch import Demands, LastMessages


class Forwarder(Thread):
    def __init__(self, forwarder_queue: "Queue[Event]", connections: Dict[int, Connection]) -> None:
        super(Forwarder, self).__init__()
        self.forwarder_queue = forwarder_queue
        self.connections = connections
        self.logger = getLogger("Forwarder")

    def run(self) -> None:
        while True:
            receiver, message = self.forwarder_queue.get()
            connection = self.connections.get(receiver)
            if connection:
                self.logger.debug("Předávání zprávy %s", receiver)
                connection.queue.put(message)
            else:
                self.logger.debug("Speciální požadavek")
                if isinstance(message, Demands):
                    self.logger.debug("Nové požadavky od routeru")
                    for connection_id, connection in self.connections.items():
                        if connection_id != message.connection_id:
                            connection.queue.put(dumps(message))
                elif isinstance(message, LastMessages):
                    self.logger.debug("Žádost o odeslání posledních zpráv")
                    for connection_id, connection in self.connections.items():
                        if connection_id != message.connection_id:
                            connection.queue.put(dumps(message))
                else:
                    self.logger.warning("Neznámý požadavek")



from typing import Optional

from edri.dataclass.event import event, response, Response
from edri.events.edri.group import Test as GroupTest


@response
class TestResponse(Response):
    random: Optional[int]

@event
class Test(GroupTest):
    pass



from random import choice
from string import ascii_letters, digits
from typing import TypeVar, Dict, Generic, Optional, LiteralString, Sized
from typing_extensions import LiteralString

T = TypeVar("T")


class Storage(Generic[T], Dict[str, T]):
    """
    A generic storage class that extends a dictionary to hold items of any type,
    identified by unique string keys. It provides functionality to automatically
    generate unique keys for items if not provided.

    Type Parameters:
        T: The type of the items to be stored in the dictionary.

    Methods:
        __init__(): Initializes a new instance of the storage.
        _get_unique_key(): Generates a unique key for storing an item.
        append(item: T, key: str = None): Adds an item to the storage with an optional key.
    """
    def __init__(self, choices: LiteralString = ascii_letters + digits, length: int = 16) -> None:
        """
        Initializes the storage with an empty dictionary.
        """
        super().__init__()
        self.choices = choices
        self.length = length

    def _get_unique_key(self) -> str:
        """
        Generates a unique key composed of random letters and digits.

        Returns:
            A unique string key of 16 characters.
        """
        while True:
            key = ''.join(choice(self.choices) for _ in range(self.length))
            if key not in self:
                return key

    def append(self, item: T, key: Optional[str] = None) -> str:
        """
        Adds an item to the storage with an automatically generated or specified key.
        If the key already exists in the storage, a KeyError is raised.

        Parameters:
            item: The item to be stored.
            key (Optional[str]): The key under which to store the item. If None, a unique key
                                 is generated automatically.

        Returns:
            The key used to store the item in the dictionary.

        Raises:
            KeyError: If the specified key already exists in the storage.
        """
        if key is None:
            key = self._get_unique_key()
        elif key in self:
            raise KeyError("Key already exists")
        self[key] = item
        return key



from os import getenv
from typing import Optional

HEALTH_CHECK_TIMEOUT = int(getenv("EDRI_HEALTH_CHECK_TIMEOUT", 10))
TOKEN_LENGTH = int(getenv("EDRI_TOKEN_LENGTH", 64))
REST_RESPONSE_TIMEOUT = int(getenv("EDRI_REST_RESPONSE_TIMEOUT", 60))

SWITCH_KEY_LENGTH = int(getenv("EDRI_SWITCH_KEY_LENGTH ", 8))
SWITCH_HOST = getenv("EDRI_SWITCH_HOST")
SWITCH_PORT = getenv("EDRI_SWITCH_POST")

UPLOAD_FILES_PREFIX = getenv("EDRI_UPLOAD_FILES_PREFIX", "edri_")
UPLOAD_FILES_KEEP_DAYS = int(getenv("EDRI_FILES_KEEP_DAYS", 0))
UPLOAD_FILES_PATH = getenv("EDRI_FILES_PATH", "/tmp/edri")

CACHE_TIMEOUT = int(getenv("EDRI_CACHE_TIMEOUT", 30))
CACHE_INFO_MESSAGE = int(getenv("EDRI_CACHE_INFO_MESSAGE", 60))

HOST = getenv("EDRI_HOST", "localhost")
ws_port_temp = getenv("EDRI_WS_PORT")
WS_PORT: Optional[int] = None
if ws_port_temp:
    WS_PORT = int(ws_port_temp)
else:
    WS_PORT = 8877

rest_port_temp = getenv("EDRI_REST_PORT")
REST_PORT: Optional[int] = None
if rest_port_temp:
    REST_PORT = int(rest_port_temp)
else:
    REST_PORT = 8878




from typing import Dict

from edri.dataclass.event import event
from edri.events.edri.group import Switch


@event
class LastMessages(Switch):
    connection_id: int
    last_messages: Dict[int, str]



from multiprocessing.connection import Connection
from typing import Type

from edri.dataclass.event import Event, event
from edri.events.edri.group import Router


@event
class Unsubscribe(Router):
    message: Type[Event]
    pipe: Connection
    request: bool



from socket import socket, AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR
from logging import getLogger
from typing import Dict
from queue import Queue
from pickle import dumps

from edri.dataclass.switch import Connection
from edri.events.edri.switch import Demands
from edri.switch import Forwarder, Receiver, Sender
from edri.config.setting import SWITCH_HOST, SWITCH_PORT


class Switch:
    def __init__(self) -> None:
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.logger = getLogger(__name__)
        self.connections: Dict[int, Connection] = {}
        self.forwarder_queue = Queue()
        self.forwarder: Forwarder = Forwarder(self.forwarder_queue, self.connections)
        self.forwarder.start()

    def handle_new_connection(self, connection: socket) -> None:
        connection_id_bytes = connection.recv(8)
        connection_id = int.from_bytes(connection_id_bytes, "big", signed=False)
        connection_info = self.connections.get(connection_id)
        self.logger.info("Nové spojení - %s", connection_id)
        if connection_info:
            self.logger.info("Obnova spojení s %s", connection_id)
            connection_info.socket = connection
            receiver = Receiver(self.forwarder_queue, connection_id, connection_info)
            receiver.start()
            sender = Sender(connection_info.queue, connection, connection_id)
            sender.start()
            self.logger.debug("Spuštěn receiver a sender")
        else:
            self.logger.info("Nové spojení s %s", connection.getsockname())
            connection_info = Connection(Queue(), connection, None)
            self.connections[connection_id] = connection_info
            receiver = Receiver(self.forwarder_queue, connection_id, connection_info)
            receiver.start()
            sender = Sender(connection_info.queue, connection, connection_id)
            sender.start()
            self.logger.debug("Spuštěn receiver a sender")

        for connection_id_demand, connection_demand in self.connections.items():
            if connection_id != connection_id_demand:
                demand = dumps(Demands(connection_id=connection_id_demand, demands=connection_demand.demands))
                self.forwarder_queue.put((connection_id, demand))

    def run(self) -> None:
        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.socket.bind((SWITCH_HOST, SWITCH_PORT))
        self.socket.listen()
        while True:
            connection, address = self.socket.accept()
            self.logger.debug("Nové spojení s %s", address)
            self.handle_new_connection(connection)


if __name__ == "__main__":
    switch = Switch()
    switch.run()



from enum import Enum, IntEnum, unique

SWITCH_NEW_DEMANDS = (18446744073709551615).to_bytes(8, "big", signed=False)
SWITCH_LAST_MESSAGES = (18446744073709551614).to_bytes(8, "big", signed=False)

SCHEDULER_TIMEOUT_MAX = 2147483  # INT_MAX bez posledních tří čísel
STREAM_CLOSE_MARK = "#&@"


@unique
class ApiType(IntEnum):
    WS = 0
    REST = 1



from dataclasses import dataclass
from datetime import datetime, timedelta
from logging import getLogger
from multiprocessing import Queue
from multiprocessing.connection import Connection
from typing import Optional, Dict

from edri.abstract import ManagerBase
from edri.config.setting import HEALTH_CHECK_TIMEOUT
from edri.dataclass.event.response import ResponseStatus
from edri.events.edri.router import HealthCheck
from edri.events.edri.scheduler import Set as SchedulerSet
from edri.events.edri.store import Set as StoreSet


class HealthChecker:
    """
    Monitors and records the health status of components within the system, scheduling regular
    health checks and storing the results for analysis and system management.

    Attributes:
        router_queue (Queue): The messaging queue used to communicate with the scheduler and store.
        logger (Logger): Logger instance for logging health check operations and outcomes.
        state (Dict[str, Record]): A dictionary storing the health status records for each component.
        last_check (Optional[datetime]): The timestamp of the last initiated health check cycle.
        components (set[ManagerBase]): A set of component instances that are monitored by the HealthChecker.

    Methods:
        set_task(): Schedules the next health check operation using the system's scheduler.
        component_add(name: str, pipe: Connection): Adds a new component to the monitoring list.
        control_start(): Initiates a health check cycle, saving the current status of components.
        control_result(name: str, status: ResponseStatus): Records the result of a health check for a component.
        save_status(): Stores the current health status of all components in the system's store.
        restart_component(pipe: Connection, time: datetime): Attempts to restart a component based on its connection.
    """
    @dataclass
    class Record:
        """
        Represents the health status of a system component.

        Attributes:
            name (str): The name of the component.
            pipe (Connection): A connection object used for communicating with the component.
            definition (Optional[ManagerBase]): The class definition of the component, if available.
            timestamp (datetime): The timestamp of the last health check.
            status (Optional[int]): The health status returned by the component, typically aligning
                                    with HTTP status codes or custom application-defined codes.
            message (Optional[str]): Additional message or information about the component's health.
        """
        name: str
        pipe: Connection
        definition: Optional[ManagerBase] = None
        timestamp: datetime = datetime.now()
        status: Optional[int] = None
        message: Optional[str] = None

    def __init__(self, router_queue: "Queue[Event]", components: set[ManagerBase]) -> None:
        """
        Initializes the HealthChecker with a set of components to monitor and the queue for communication.
        """
        self.router_queue = router_queue
        self.logger = getLogger(__name__)
        self.state: Dict[str, HealthChecker.Record] = {}
        self.last_check: Optional[datetime] = None
        self.components = components
        self.set_task()

    def set_task(self) -> None:
        """
        Schedules the health check operation at a regular interval defined by HEALTH_CHECK_TIMEOUT.
        """
        scheduler_set = SchedulerSet(
            event=HealthCheck(),
            when=datetime.now() + timedelta(seconds=HEALTH_CHECK_TIMEOUT),
            repeat=timedelta(seconds=HEALTH_CHECK_TIMEOUT),
            identifier="HealthCheckerTask"
        )
        self.router_queue.put(scheduler_set)

    def component_add(self, name: str, pipe: Connection) -> None:
        """
        Registers a new component for health monitoring.
        """
        if name not in self.state:
            component = None
            for component in self.components:
                if component.name == name:
                    break
            self.state[name] = self.Record(name, pipe, component)
            self.logger.debug("Manager was added %s %s", name, pipe)

    def control_start(self) -> None:
        """
        Marks the start of a health check cycle, updating records and preparing for result collection.
        """
        self.save_status()
        self.last_check = datetime.now()

    def control_result(self, name: str, status: ResponseStatus) -> None:
        """
        Records the health check result for a specific component.
        """
        self.logger.debug("Add status record to %s - %s", name, status)
        record = self.state[name]
        record.timestamp = datetime.now()
        record.status = status.value

    def save_status(self) -> None:
        """
        Compiles and stores the health status of all monitored components.
        """
        store_set = StoreSet(name="health_check", value={})
        now = datetime.now()
        store_set.value["timestamp"] = now.isoformat()
        for name, record in self.state.items():
            if record.status is None:
                self.logger.debug("%s has not been checked yet", record.name)
                continue
            if (record.timestamp + timedelta(seconds=HEALTH_CHECK_TIMEOUT * 2)) < now:
                self.logger.warning("%s did not send a status message - last status %s - %s",
                                    record.name, record.status, record.message)
            else:
                store_set.value[name] = {
                    "status": record.status,
                    "message": record.message,
                }
        self.router_queue.put(store_set)

    def restart_component(self, pipe: Connection, time: datetime) -> None:
        """
        Attempts to restart a component that is not responding or is reported to have issues.
        """
        self.logger.warning("Restarting component")
        for name, component in self.state.items():
            if component.pipe == pipe:
                if component.definition:
                    component.definition.__class__(router_queue=self.router_queue, from_time=time).start()
                    self.logger.info("Component restarted")
                else:
                    self.logger.error("Component was found - definition was missing")
                del self.state[name]
                break
        else:
            self.logger.error("Component was not found")



from multiprocessing import Queue
from multiprocessing.connection import wait, Connection
from time import sleep
from typing import Optional, Set, cast
from logging import getLogger

from edri.abstract import ManagerBase
from edri.config.constant import ApiType
from edri.dataclass.client import Client
from edri.dataclass.event import Event, api_events
from edri.dataclass.event.response import ResponseStatus
from edri.events.api import group, manage, client
from edri.utility.storage import Storage


class ApiBroker(ManagerBase):
    """
        The ApiBroker class serves as a central component for managing API interactions,
        client registrations, and event dispatching in a system that utilizes
        multiprocessing. It inherits from ManagerBase to leverage common management
        functionalities and extends it with API-specific capabilities.

        Attributes:
            api_broker_queue (Queue): A multiprocessing queue used to receive events
            specifically for the API broker.
            clients (Optional[Storage[Client]]): A storage mechanism for registered
            clients, allowing for quick access and management.
            db: Placeholder attribute for potential database connections or operations,
            not implemented by default.
            events (dict): A mapping of event resources to event classes, derived from
            the global api_events definition.

        Methods:
            __init__(router_queue: "Queue[Event]", api_broker_queue: "Queue[Event]"): Constructor for the
            ApiBroker class.
            name (property): Returns the name of the ApiBroker instance.
            send_specific(event: Event): Sends an event to a specific client based on
            the event's key.
            send_all(event: Event): Broadcasts an event to all clients of a specific
            API type.
            client_register(event: client.Register): Registers a new client and sends
            a confirmation event.
            client_unregister(event: Event): Unregisters a client based on the event's
            key.
            event_register(event: manage.Register): Registers a client for specific
            events and parameters.
            solve(event: Event): Determines the appropriate method to handle an incoming
            event.
            _prepare(): Prepares the ApiBroker for operation by initializing storage for
            clients.
            additional_pipes (property): Returns a set of additional pipes to monitor
            for incoming events.
            run_resolver(): Main loop for processing incoming events and dispatching them
            to the appropriate handlers.
            resolve_callback(event: Event, pipe: Pipe): Callback method for handling
            events received from clients.

        Usage:
            This class should be instantiated with the appropriate queues and used as
            part of a larger multiprocessing application to manage API interactions and
            client communications. It requires subclassing ManagerBase and implementing
            any additional abstract methods or properties as needed.
        """
    def __init__(self, router_queue: "Queue[Event]", api_broker_queue: "Queue[Event]") -> None:
        """
        Initializes an ApiBroker instance with specific queues for routing and
        API broker events.

        Parameters:
            router_queue (Queue): The queue for receiving routed events from other
            components of the system.
            api_broker_queue (Queue): The queue specifically for the API broker to
            receive events.
        """
        super().__init__(router_queue)
        self.api_broker_queue = api_broker_queue
        self.logger = getLogger(__name__)
        self.clients: Storage[Client]
        self.db = None
        self.events = {event.resource: event.event for event in api_events}

    @property
    def name(self) -> str:
        """
        Property returning the name of the ApiBroker instance. Primarily used for
        logging and identification purposes.

        Returns:
            str: The name "ApiBroker".
        """
        return "ApiBroker"

    def _send(self, client: Client, event: Event) -> None:
        """
        Attempts to send an event to a specific client. If the connection is broken
        or reset during sending, it logs a warning and unregisters the client.

        Parameters:
            client (Client): The client object representing the target for the event.
            event (Event): The event to be sent to the client.

        Returns:
            None
        """
        try:
            client.socket.send(event)
        except BrokenPipeError as e:
            self.logger.warning("Cannot be send %s", client.socket, exc_info=e)
            self.client_unregister(event)
        except ConnectionResetError as e:
            self.logger.warning("Cannot be send %s", client.socket, exc_info=e)
            self.client_unregister(event)

    def send_specific(self, event: Event) -> None:
        """
        Sends an event to the specific client identified by the event's key. Logs
        a warning if the client is not found.

        Parameters:
            event (Event): The event containing the key of the target client.

        Returns:
            None
        """
        self.logger.debug("-> %s", event)
        if not event._key:
            self.logger.error("Key not found in the event %s", event)
            return
        try:
            client = self.clients[event._key]
        except KeyError:
            self.logger.warning("Client was not found: %s", event)
            return
        self._send(client, event)

    def send_all(self, event: Event) -> None:
        """
         Broadcasts an event to all clients that are of a specific API type (e.g., WebSocket).
         It iterates through all clients and sends the event to those matching the type.

         Parameters:
             event (Event): The event to be broadcasted.

         Returns:
             None
         """
        self.logger.debug("->* %s", event)
        for client in self.clients.values():
            if client.type == ApiType.WS:
                self._send(client, event)

    ##### API #####
    def client_register(self, event: client.Register) -> None:
        """
        Handles the registration of a new client based on the received register event.
        It assigns a unique key to the client, stores it, and acknowledges the registration.

        Parameters:
            event (client.Register): The event containing the client's registration information.

        Returns:
            None
        """
        event._key = self.clients.append(Client(event.socket, ApiType.WS))
        event.socket = None
        event.response.set_status(ResponseStatus.OK)
        self.send_specific(event)

    def client_unregister(self, event: Event) -> None:
        """
        Handles the unregistration of a client, removing them from the stored clients
        based on the event's key. It performs cleanup if necessary.

        Parameters:
            event (Event): The event containing the key of the client to be unregistered.

        Returns:
            None
        """
        try:
            if event._key:
                del self.clients[event._key]
            else:
                self.logger.warning("Client cant be unregistered because key is missing!")
        except KeyError:
            pass

    def event_register(self, event: manage.Register) -> None:
        """
          Registers a client for specific events and their parameters, allowing for
          customized event handling. This method updates the client's event subscriptions
          and their associated parameters based on the provided event registration details.

          Parameters:
              event (manage.Register): An object representing the event registration request.
              This object must include a non-empty `_key` attribute identifying the client,
              a list of `events` the client wishes to subscribe to, and corresponding
              `parameters` and `values` for those events.

          Returns:
              None

          Raises:
              KeyError: If the event `_key` attribute is missing or if the `_key` does not
              correspond to any existing client in the `clients` dictionary.

              ValueError: If the `events` list in the `event` object is empty, indicating
              that no events have been specified for registration.

              LookupError: If any of the events specified in the `event.events` list do not
              exist in the `self.events` mapping, indicating an attempt to register for
              undefined events.

          This method first verifies that the event `_key` is present and corresponds to an
          existing client. It then proceeds to update the client's event subscriptions and
          parameters. If successful, the client's response status is set to `OK`, and the
          updated event registration details are sent back to the client via their socket.
          """
        if not event._key:
            raise KeyError("Event key is missing!")
        client = self.clients[event._key]
        client.events = {self.events[event] for event in event.events}
        client.parameters = dict(zip(event.parameters, event.values))
        event.response.set_status(ResponseStatus.OK)
        client.socket.send(event)

    def solve(self, event: Event) -> None:
        """
        Determines how to handle an incoming event, directing it to specific clients
        or broadcasting it based on the event's properties.

        Parameters:
            event (Event): The event to be handled.

        Returns:
            None
        """
        if event._key:
            self.send_specific(event)
        else:
            found = False
            for client in self.clients.values():
                if event.__class__ in client.events:
                    found = True
                    client.socket.send(event)
            if not found:
                self.send_all(event)

    def _prepare(self) -> None:
        """
         Prepares the ApiBroker for operation by initializing necessary components,
         such as client storage.

         Returns:
             None
         """
        super()._prepare()
        self.clients = Storage()

    @property
    def additional_pipes(self) -> Set[Connection]:
        """
        Provides a set of additional pipes that should be monitored for incoming
        events, specifically the API broker queue.

        Returns:
            Set[Pipe]: A set containing the reader end of the API broker queue.
        """
        return {self.api_broker_queue._reader}

    def run_resolver(self) -> None:
        """
        The main event loop of the ApiBroker, continuously checking for and processing
        incoming events from various sources.

        Returns:
            None
        """
        while True:
            pipes: set[Connection] = set()
            if self.router_pipe:
                pipes.add(self.router_pipe)
            if self._workers:
                pipes.update(worker.pipe for worker in self._workers.values())
            pipes.update(self.additional_pipes)
            if not pipes:
                sleep(1)
                continue
            try:
                active_pipes = cast(list[Connection], wait(pipes, timeout=10))
                for active_pipe in active_pipes:
                    if active_pipe == self.router_pipe:
                        event: Event = self.router_pipe.recv()
                        self.logger.debug("Received event: %s", event)
                        self.resolve(event)
                    else:
                        try:
                            event = active_pipe.recv()
                        except EOFError as e:
                            self.logger.error("Communication problems", exc_info=e)
                            continue
                        except OSError as e:
                            self.logger.error("OS problems", exc_info=e)
                            continue
                        self.resolve_callback(event, active_pipe)

            except KeyboardInterrupt:
                return

    def resolve_callback(self, event: Event, pipe: Connection) -> None:
        """
        Handles incoming events received from clients or other components, dispatching
        them to the appropriate handling method based on their type.

        Parameters:
            event (Event): The incoming event to be handled.
            pipe (Connection): The communication pipe from which the event was received.

        Returns:
            None
        """
        self.logger.debug("<- Client %s", event)
        if isinstance(event, group.Client):
            if isinstance(event, client.Register):
                self.client_register(event)
            elif isinstance(event, client.Unregister):
                self.client_unregister(event)
        elif isinstance(event, group.Manage):
            if isinstance(event, manage.Register):
                self.event_register(event)
        else:
            self.logger.debug("Router <- API Broker %s", event)
            self.router_queue.put(event)

    def _prepare_resolvers(self) -> None:
        """
        Prepares the resolvers for handling specific types of events, enhancing
        the base implementation with API-specific event handling.

        Returns:
            None
        """
        super()._prepare_resolvers()
        for event in api_events:
            if "response" not in event.event.__annotations__:
                self._requests[event.event] = self.solve
            else:
                self._responses[event.event] = self.solve



from multiprocessing.connection import Connection
from dataclasses import dataclass, field
from typing import Type

from edri.dataclass.event import Event
from edri.config.constant import ApiType


@dataclass
class Client:
    socket: Connection
    type: ApiType
    events: set[Type[Event]] = field(default_factory=set)
    parameters: dict[str, str] = field(default_factory=dict)



from dataclasses import _process_class, dataclass, _FIELD_BASE, field
from enum import Enum
from inspect import isclass
from pathlib import Path
from types import NoneType
from typing import Type, Union
from typeguard import check_type
from sys import version_info

from edri.config.constant import ApiType
from edri.dataclass.event.response import Response
from edri.dataclass.event.event import Event, BaseIdentification
from edri.utility.function import camel2snake


@dataclass
class ApiEventSharing(Enum):
    SPECIFIC = "SPECIFIC"
    ALL = "ALL"


class Method(Enum):
    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


@dataclass
class ApiEvent:
    url: str
    resource: str
    sharing: ApiEventSharing
    event: Type[Event]
    exclude: list[ApiType]


Methods = tuple[Method] | Method

events: list[Event] = list()
api_events: list[ApiEvent] = list()
allowed_types = (str, int, float, bool, Methods, Path)


def api(cls=None, /, *, init=True, repr=True, eq=True, order=False,
        unsafe_hash=False, frozen=False, match_args=True,
        kw_only=False, slots=False, url=None, resource=None, sharing=None, exclude=None, weakref_slot=None):
    def wrapper(cls):
        dataclass = _event(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                           frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots,
                           weakref_slot=weakref_slot)
        for name, field in dataclass.__dataclass_fields__.items():
            if name.startswith("_"):
                continue
            if name == "method":
                check_type(field.default, Methods)
            elif name == "response":
                pass
            elif field.type not in allowed_types and isclass(field.type) and not issubclass(field.type, Enum):
                item_type = getattr(field.type, "__origin__", None)
                item_args = getattr(field.type, "__args__", None)
                if item_type == Union:
                    position = 0 if item_args.index(NoneType) == 1 else 1
                    item_type = item_args[position]
                    if item_type in (list, tuple):
                        item_args = item_type.__args__
                if item_type in (list, tuple):
                    if len(item_args) > 1:
                        raise TypeError("Only one child type is allowed got %s" % len(item_args))
                    elif item_args[0] not in allowed_types and not hasattr(item_args[0], "fromisoformat"):
                        raise TypeError("%s cannot be used as a type for API event" % item_args[0])
                elif item_type not in allowed_types and not hasattr(field.type, "fromisoformat"):
                    raise TypeError("%s cannot be used as a type for API event" % field.type)

        if not hasattr(dataclass, "method"):
            if exclude is None or ApiType.REST not in exclude:
                raise ValueError("Method is required for REST")
        else:
            if not isinstance(dataclass.method, tuple):
                dataclass.method = (dataclass.method,)
        api_events.append(ApiEvent(
            url if url else f"/{camel2snake(dataclass.__name__)}",
            resource if resource else camel2snake(dataclass.__name__).replace("_", "-"),
            sharing if sharing else ApiEventSharing.SPECIFIC,
            dataclass,
            exclude if exclude is not None else []))
        return dataclass

    if cls is None:
        return wrapper

    return wrapper(cls)


def _event(cls, init, repr, eq, order, unsafe_hash, frozen, match_args, kw_only, slots, weakref_slot) -> object:
    response = cls.__annotations__.get("response", False)
    if response:
        attribute = getattr(cls, "response", False)
        if attribute:
            attribute.init = False
            attribute.default_factory = response
        else:
            setattr(cls, "response", field(init=False, default_factory=response))
    if version_info < (3, 11, 0):
        dataclass = _process_class(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                   frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots)
    else:
        dataclass = _process_class(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                   frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots,
                                   weakref_slot=weakref_slot)

    return dataclass


def event(cls=None, /, *, init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
          match_args=True, kw_only=False, slots=False, weakref_slot=False):
    def wrapper(cls):
        return _event(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                      frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots, weakref_slot=weakref_slot)

    if cls is None:
        return wrapper

    return wrapper(cls)


def _response(cls, init, repr, eq, order, unsafe_hash, frozen, match_args, kw_only, slots, weakref_slot):
    annotations = {}
    for name, value in cls.__annotations__.items():
        annotations[name] = value | None
        attribute = getattr(cls, name, False)
        if attribute:
            setattr(cls, name, field(init=False, default=attribute))
        else:
            setattr(cls, name, field(init=False, default=None))

    if version_info < (3, 11, 0):
        dataclass = _process_class(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                   frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots)
    else:
        dataclass = _process_class(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                                   frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots,
                                   weakref_slot=weakref_slot)

    return dataclass


def response(cls=None, /, *, init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
             match_args=True, kw_only=False, slots=False, weakref_slot=None):
    def wrapper(cls):
        return _response(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash,
                         frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots, weakref_slot=weakref_slot)

    if cls is None:
        return wrapper

    return wrapper(cls)


__all__ = ["Event", "event", "Response", "response"]



from datetime import datetime
from multiprocessing.connection import Connection
from typing import Type, Optional

from edri.dataclass.event import Event, Response, event, response
from edri.events.edri.group import Router


@response
class SubscribeResponse(Response):
    pass


@event
class Subscribe(Router):
    event_type: Type[Event]
    name: str
    pipe: Optional[Connection]
    request: bool
    from_time: datetime
    response: SubscribeResponse


from multiprocessing.connection import Connection
from typing import Optional

from edri.dataclass.event import event
from edri.events.edri.group import Router


@event
class SubscribedAll(Router):
    pipe: Optional[Connection]



from pickle import loads
from threading import Thread
from queue import Queue
from socket import SHUT_RDWR
from logging import getLogger

from edri.config.constant import SWITCH_NEW_DEMANDS, SWITCH_LAST_MESSAGES
from edri.dataclass.switch import Connection
from edri.events.edri.switch import Demands, LastMessages


class Receiver(Thread):
    def __init__(self, forwarder_queue: "Queue[Event]", connection_id: int, connection: Connection) -> None:
        super(Receiver, self).__init__()
        self.forwarder_queue = forwarder_queue
        self.connection = connection
        self.connection_id = connection_id
        self.logger = getLogger(f"Receiver {connection_id}")

    def run(self) -> None:
        try:
            while True:
                try:
                    recipient_bytes = self.connection.socket.recv(8)
                except ConnectionResetError:
                    self.logger.error("Klient ukončil spojení", )
                    return
                if not recipient_bytes:
                    self.connection.socket.shutdown(SHUT_RDWR)
                    self.logger.error("Ukončeno spojení se switchem", )
                    return
                recipient = int.from_bytes(recipient_bytes, "big", signed=False)
                self.logger.debug("Nová zpráva pro %s", recipient)

                if recipient_bytes == SWITCH_NEW_DEMANDS:
                    self.logger.debug("Nové požadavky")
                    demands_size_bytes = self.connection.socket.recv(8)
                    demands_size = int.from_bytes(demands_size_bytes, "big", signed=False)

                    demands_bytes = self.connection.socket.recv(demands_size)
                    demands = loads(demands_bytes)
                    self.connection.demands = demands
                    self.logger.debug("Požadavky %s", demands)
                    demands_switch = Demands(connection_id=self.connection_id, demands=demands)
                    self.forwarder_queue.put((0, demands_switch))
                elif recipient_bytes == SWITCH_LAST_MESSAGES:
                    self.logger.debug("Poslední zprávy")
                    last_messages_size_bytes = self.connection.socket.recv(8)
                    demands_size = int.from_bytes(last_messages_size_bytes, "big", signed=False)

                    last_messages_bytes = self.connection.socket.recv(demands_size)
                    last_messages = loads(last_messages_bytes)

                    last_messages_switch = LastMessages(connection_id=self.connection_id,
                                                        last_messages=last_messages.get_response().data["last_messages"])
                    self.forwarder_queue.put((0, last_messages_switch))

                else:
                    message_size_bytes = self.connection.socket.recv(8)
                    message_size = int.from_bytes(message_size_bytes, "big", signed=False)

                    message_bytes = self.connection.socket.recv(message_size)

                    self.logger.debug("přijato od %s %sb", recipient, len(message_bytes))
                    self.forwarder_queue.put((recipient, message_bytes))
        except Exception as e:
            self.logger.error("Neznámá chyba", exc_info=e)



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Optional, Union, List


class ETA:
    # @dataclass
    # class ETARecord:
    #     progress: float
    #     delta: timedelta

    history_count = 10

    class Form:
        @staticmethod
        def seconds(quantity: int) -> str:
            if quantity == 1:
                return "sekunda"
            elif 2 <= quantity <= 4:
                return "sekundy"
            else:
                return "sekund"

    def __init__(self) -> None:
        # self.records: list[ETA.ETARecord] = []
        self.start = datetime.now()
        self.minimum: float = 0.0
        self.maximum: float = 0.0
        self.average: float = 0.0
        self.average_count: int = 0
        self.variance_list: List[float] = []

    def add_record(self, progress: float) -> None:
        current_average = progress / (datetime.now() - self.start).total_seconds()

        self.variance_list.append(abs(current_average - self.average))
        if len(self.variance_list) > self.history_count:
            self.variance_list.pop(0)

        current_average = (self.average * self.average_count + current_average)
        self.average_count += 1

        try:
            self.average = current_average / self.average_count
        except ZeroDivisionError:
            self.average = current_average

    def get_estimation(self) -> Union[str, None]:
        elapsed = datetime.now() - self.start
        remaining = (1 / self.average) - elapsed.total_seconds()
        variance = (sum(self.variance_list) / len(self.variance_list)) * 100
        variance = remaining * variance  # max(0.05, variance)
        return f"Zbývá: {remaining} +- {variance}"



from dataclasses import dataclass
from queue import Queue
from typing import Optional
from socket import socket

from edri.dataclass.event import Event

from edri.events.edri.router import Demands


@dataclass
class Connection:
    queue: Queue[Event]
    socket: socket
    demands: Optional[Demands]



from typing import Optional


def camel2snake(title: str) -> str:
    """
    Converts a camelCase string to a snake_case string.

    Parameters:
        title (str): The camelCase string to be converted.

    Returns:
        str: The converted snake_case string.

    Examples:
        >>> camel2snake("camelCaseExample")
        'camel_case_example'
        >>> camel2snake("CamelCaseExample")
        'camel_case_example'
    """
    new_title = ""
    underscore = False
    for letter in title:
        if 65 <= ord(letter) <= 90:
            if underscore:
                new_title += '_'
                underscore = False
            new_title += chr(ord(letter) + 32)
        else:
            underscore = True
            new_title += letter
    return new_title


def snake2camel(title: str, exceptions: Optional[tuple[str]] = None) -> str:
    """
    Converts a snake_case string to a camelCase string, with optional exceptions
    where the conversion should not be applied.

    Parameters:
        title (str): The snake_case string to be converted.
        exceptions (Optional[tuple[str]]): A tuple of strings that, when matched at the
                                      beginning of the title, prevent conversion and
                                      return the title as is. Defaults to None.

    Returns:
        str: The converted camelCase string if no exception is matched; otherwise,
             the original string.

    Examples:
        >>> snake2camel("snake_case_example")
        'snakeCaseExample'
        >>> snake2camel("id_example", exceptions=("id_",))
        'id_example'

    Notes:
        - The first segment of the snake_case string remains lowercase, aligning
          with the camelCase convention.
        - If the title starts with any of the specified exceptions, it is returned
          unchanged, useful for preserving certain identifiers.
    """
    if exceptions is None:
        exceptions = ("id_",)
    if title.startswith(exceptions):
        return title
    title_list = title.split('_')
    title_list = [title_list[0]] + [word[0].upper() + word[1:] for word in title_list][1:]
    return ''.join(title_list)



from dataclasses import dataclass, field
from random import choices
from string import ascii_letters, digits
from typing import Optional, TypeGuard, TYPE_CHECKING
from zlib import adler32

from edri.config.setting import SWITCH_KEY_LENGTH
from edri.dataclass.event import Response
from edri.dataclass.event.response import ResponseStatus

if TYPE_CHECKING:
    from typing import TypeVar

    _T = TypeVar("_T")

    class _ResponseExists:
        response: Response


@dataclass
class BaseIdentification:
    id: Optional[int] = None
    key: str = field(default_factory=lambda: "".join(choices(ascii_letters + digits, k=SWITCH_KEY_LENGTH)))
    received: bool = False

    def __repr__(self) -> str:
        return f"{self.id}:{self.key}"


@dataclass
class Event:
    _key: Optional[str] = field(init=False, default=None)
    _stream: Optional[str] = field(init=False, default=None)
    _switch: Optional[BaseIdentification] = field(init=False, default=None)
    _headers: list[tuple[bytes, bytes]] | None = field(init=False, default=None)
    _cookies: dict[str, str] | None = field(init=False, default=None)
    _worker: str | None = field(init=False, default=None)
    response: Optional[Response] = field(init=False, default=None)

    def get_response(self) -> Optional[Response]:
        if self.response and self.response.get_status() != ResponseStatus.NONE:
            return self.response
        return None

    def has_response(self) -> bool:
        return self.response is not None and self.response.get_status() != ResponseStatus.NONE

    def set_response(self, response: Response) -> None:
        self._switch = None
        self.response = response

    def remove_response(self) -> None:
        self._switch = None
        self.response = None

    @classmethod
    def hash_name(cls) -> int:
        base = cls.__class__
        return adler32(f"{base.__module__}.{base.__qualname__}".encode())



from multiprocessing import Queue

from edri.abstract import ManagerBase
from edri.utility.connector import Connector
from edri.utility.scheduler import Scheduler, Job
from edri.config.setting import TOKEN_LENGTH
from edri.dataclass.event import Event
from edri.router import Router
from edri.rest import REST
from edri.utility.store import Store
from edri.utility.api_broker import ApiBroker
from edri.websocket import API as WebSocket


class EDRI:
    def __init__(self) -> None:
        self.router_queue: "Queue[Event]" = Queue()
        self.components: set[ManagerBase] = set()
        self.store: Store
        self.scheduler: Scheduler
        self.rest: REST
        self.websocket: WebSocket
        self.api_broker: ApiBroker
        self.connector: Connector
        self.router: Router

    def add_component(self, component: ManagerBase) -> None:
        self.components.add(component)

    def start_store(self) -> None:
        self.store = Store(self.router_queue)
        self.store.start()

    def start_scheduler(self, jobs: list[Job] | None = None) -> None:
        if not jobs:
            jobs = []
        self.scheduler = Scheduler(self.router_queue, jobs)
        self.scheduler.start()

    def api_broker_start(self) -> None:
        self.api_broker = ApiBroker(self.router_queue, Queue())
        self.api_broker.start()

    def start_rest(self) -> None:
        if not hasattr(self, "api_broker"):
            self.api_broker_start()
        self.rest = REST(self.api_broker.api_broker_queue)
        self.rest.start()

    def start_websocket(self) -> None:
        if not hasattr(self, "api_broker"):
            self.api_broker_start()
        self.websocket = WebSocket(self.api_broker.api_broker_queue)
        self.websocket.start()

    def start_connector(self) -> None:
        self.connector = Connector(self.router_queue)
        self.connector.start()

    def run(self) -> None:
        self.router = Router(self.router_queue, self.components)
        for component in self.components:
            component.start(self.router_queue)
        try:
            self.router.run()
        except KeyboardInterrupt:
            self.router.quit()



from dataclasses import dataclass
from datetime import datetime, timedelta
from logging import getLogger
from multiprocessing.queues import Queue
from typing import Union, Optional

from edri.abstract import ManagerBase
from edri.dataclass.event import Event
from edri.dataclass.event.response import ResponseStatus
from edri.events.edri.scheduler import Set, Update, Cancel
from edri.config.constant import SCHEDULER_TIMEOUT_MAX
from edri.utility.storage import Storage


@dataclass
class Job:
    """
    Represents a scheduled task within the system.

    Attributes:
        event (Event): The event to be triggered when the job is executed.
        when (datetime): The specific time at which the job should be executed.
        repeat (Union[timedelta, bool]): Indicates how often the job should repeat.
            If False, the job does not repeat. If a timedelta, it repeats at the interval
            specified by the timedelta.
    """
    event: Event
    when: datetime
    repeat: Optional[timedelta] = None


class Scheduler(ManagerBase):
    """
    A scheduler for managing and executing timed tasks based on the Job dataclass.
    Inherits from ManagerBase to utilize a common management infrastructure for
    event processing.

    Attributes:
        jobs (Storage[Job]): A storage for managing jobs. Initialized as None and
        should be set to a Storage instance before starting the scheduler.
        initial_jobs (list[Job]): A list of jobs to be initially loaded into the
        scheduler upon start.

    Methods:
        after_start(): Initializes the jobs storage and loads initial jobs.
        solve_req_set(event: Set): Handles requests to set (schedule) new jobs.
        solve_req_update(event: Update): Handles requests to update existing jobs.
        solve_req_cancel(event: Cancel): Handles requests to cancel existing jobs.
        get_next_job(): Retrieves the next job to be executed, based on its scheduled time.
        run_pending(): Executes pending jobs based on the current time and schedules
        repeat executions if necessary.
        run_resolver(): Continuously checks for and executes pending jobs, and listens
        for incoming job management requests.
    """
    def __init__(self, router_queue: Queue[Event], jobs: list[Job]) -> None:
        """
         Initializes the Scheduler instance with a router queue and an initial list
         of jobs.

         Parameters:
             router_queue (Queue): The multiprocessing queue for receiving and routing
             event messages.
             jobs (list[Job]): An initial list of jobs to be scheduled upon startup.
         """
        super().__init__(router_queue, getLogger(__name__))
        self.jobs: Storage[Job]
        self.initial_jobs = jobs

    def after_start(self) -> None:
        """
        Prepares the scheduler for operation by initializing the jobs storage and
        loading the initial jobs.
        """
        self.jobs = Storage()
        for job in self.initial_jobs:
            self.jobs.append(job)

    def solve_req_set(self, event: Set) -> None:
        """
        Processes a request to schedule a new job. Adds the job to the scheduler's
        storage and assigns it a unique identifier.

        Parameters:
            event (Set): An event containing the details of the job to be scheduled,
            including the event to trigger, when to trigger it, and whether it should
            repeat.

        Returns:
            None: This method does not return a value but updates the event's response
            with the identifier of the scheduled job.
        """
        try:
            key = self.jobs.append(Job(event.event, event.when, event.repeat), event.identifier)
        except KeyError as e:
            self.logger.warning("Scheduled task already exists: %s", event.identifier, exc_info=e)
            return

        event.response.identifier = key
        self.logger.debug("New scheduled task: %s", key)

    def solve_req_update(self, event: Update) -> None:
        """
        Processes a request to update an existing job. Allows updating the job's
        event, execution time, and repetition interval.

        Parameters:
            event (Update): An event containing the job's identifier and the new
            values for the event message, execution time, and/or repetition interval.

        Returns:
            None: Updates the specified job and modifies the event's response status
            accordingly.
        """
        job = self.jobs.get(event.identifier, None)
        if not job:
            event.response.set_status(ResponseStatus.FAILED)
            self.logger.debug("Scheduled task %s was not found", event.identifier)
            return
        if event.message is not None:
            job.event = event.message
        if event.when is not None:
            job.when = event.when
        if event.repeat is not None:
            job.repeat = event.repeat
        event.response.set_status(ResponseStatus.OK)
        self.logger.debug("Scheduled task %s was changed", event.identifier)

    def solve_req_cancel(self, event: Cancel) -> None:
        """
        Processes a request to cancel an existing job. Removes the job from the
        scheduler's storage.

        Parameters:
            event (Cancel): An event containing the identifier of the job to be
            cancelled.

        Returns:
            None: Removes the specified job and modifies the event's response status
            accordingly.
        """
        try:
            del self.jobs[event.identifier]
        except KeyError:
            event.response.set_status(ResponseStatus.FAILED)
            self.logger.debug("Scheduled task %s was not found", event.identifier)
            return

        event.response.set_status(ResponseStatus.OK)
        self.logger.debug("Scheduled task %s was cancelled", event.identifier)

    def get_next_job(self) -> tuple[str, Optional[Job]]:
        """
        Identifies the next job that is scheduled to be executed.

        Returns:
            Optional[tuple[str, Job]]: A tuple containing the identifier of the next
            job and the job itself, if any jobs are scheduled; None otherwise.
        """
        if not self.jobs:
            return "", None
        if len(self.jobs) == 1:
            return next(iter(self.jobs.items()))
        jobs = iter(self.jobs.items())
        key, next_job = next(jobs)
        for key, job in jobs:
            if next_job.when > job.when:
                next_job = job
        return key, next_job

    def run_pending(self) -> None:
        """
        Checks for and executes any jobs that are due to be run. If a job is set
        to repeat, it reschedules the job based on its repetition interval.

        Returns:
            None
        """
        now = datetime.now()
        for key, job in list(self.jobs.items()):
            if now >= job.when:
                self.router_queue.put(job.event)
                if job.repeat:
                    job.when = job.when + job.repeat
                else:
                    self.logger.debug("Scheduled task %s was ended", key)
                    del self.jobs[key]

    def run_resolver(self) -> None:
        """
        The main loop of the scheduler. Continuously checks for and executes pending
        jobs, listens for job management requests, and adjusts wait times dynamically
        based on the schedule.

        Returns:
            None
        """
        while True:
            try:
                self.run_pending()
                self.logger.debug("Count of scheduled task: %s", len(self.jobs))
                key, next_job = self.get_next_job()

                if not next_job:
                    timeout = SCHEDULER_TIMEOUT_MAX
                else:
                    now = datetime.now()
                    timeout = (next_job.when - now).total_seconds()
                    if timeout > SCHEDULER_TIMEOUT_MAX:
                        timeout = SCHEDULER_TIMEOUT_MAX
                    self.logger.debug("Next scheduled task %s will be executed after %ss", key, timeout)
                if self.router_pipe.poll(timeout=timeout):
                    message = self.router_pipe.recv()
                    self.logger.debug("Received message: %s", message)
                    self.resolve(message)
            except KeyboardInterrupt:
                return



from setuptools import setup, find_packages

long_description = """**EDRI (Event Driven Routing Infrastructure)** is a project aimed at enhancing application development by utilizing 
parallelism—executing code across multiple threads, processes, or even different machines. It adopts an **event-driven** strategy, 
where events act as messages exchanged between various components to control and manage operations. Drawing parallels with TCP/IP in 
computer networks, EDRI's **routing infrastructure** facilitates the exchange and delivery of events. Instead of IP ranges, event types (
e.g., file uploads) are used to determine the destination of each event, ensuring efficient and targeted delivery to the appropriate 
recipients. This approach streamlines the process of managing distributed tasks within applications."""

setup(
    name='edri',
    version='2024.06.03',
    packages=find_packages(),
    description='Event Driven Routing Infrastructure',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Marek Olšan',
    author_email='marek.olsan@gmail.com',
    install_requires=[
        "typeguard>=4.0",
        "websockets>=12.0",
        "requests>=2.0",
        "validators>=0.22.0",
        "uvicorn>=0.27.0",
        "typing_extensions>=4.0",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "Natural Language :: Czech",
        "Typing :: Typed"
    ],
)



from typing import Optional, Any

from edri.dataclass.event import Event, event, response, Response
from edri.events.edri.group import Store


@response
class GetResponse(Response):
    value: Any


@event
class Get(Store):
    name: str
    data: Optional[Event] = None
    response: GetResponse



from typing import Any

from edri.dataclass.event import event, response, Response
from edri.events.edri.group import Store


@response
class DeleteResponse(Response):
    value: Any


@event
class Delete(Store):
    name: str
    response: DeleteResponse



from abc import ABC
from multiprocessing.connection import Connection
from threading import Thread
from typing import TypeVar

from edri.dataclass.event import Event
from edri.abstract.worker import Worker

T = TypeVar('T', bound=Event)


class WorkerThread(Worker[T], Thread, ABC):
    """
    A WorkerThread class that combines the custom Worker class with threading
    capabilities, designed to run as a separate thread. This class is abstract
    and should be subclassed to implement specific worker behavior in a threaded
    environment.

    Attributes:
        Inherits all attributes from the Worker class and the Thread class.

    Methods:
        __init__(pipe: Connection, event, name: str): Initializes a new instance
        of the WorkerThread.

    Parameters:
        pipe (Connection): A connection object used for communication between
        threads. This should be an instance of multiprocessing.Connection,
        although it's used here in a threading context for inter-thread
        communication.
        event: The event type that the worker will handle. The specific type is
        determined by the EventType generic type variable, which should be
        defined in the subclass.
        name (str): A name for the thread. This is used internally by the
        threading module to identify the thread, especially useful for debugging
        and logging.

    Returns:
        None: The constructor initializes the thread but does not return a value.

    Raises:
        Inherits any exceptions raised by the Worker class or the Thread class
        constructors.

    Note:
        As this class is abstract, it cannot be instantiated directly. It requires
        subclassing and implementation of the abstract methods defined in the
        Worker class. This approach allows for the creation of specialized worker
        threads that can handle specific types of events or tasks in a multi-threaded
        application.
    """

    def __init__(self, pipe: Connection, event: T, name: str) -> None:
        Worker.__init__(self, pipe, event, name)
        Thread.__init__(self)



from edri.config.constant import ApiType
from edri.dataclass.event import api, Methods, Method, Response, response
from edri.events.api.group import Manage


@response
class RegisterResponse(Response):
    pass


@api(resource="register", exclude=[ApiType.REST])
class Register(Manage):
    events: list[str]
    parameters: list[str]
    values: list[str]
    # method: Methods = Method.PUT
    response: RegisterResponse



from abc import ABC, abstractmethod
from logging import getLogger
from multiprocessing import Pipe, Queue
from multiprocessing.connection import Connection
from typing import Tuple

from edri.events.api.client import Register, Unregister
from edri.config.constant import ApiType


class APIBase(ABC):
    """
    An abstract base class for API components in a system that uses multiprocessing
    for communication between different parts of the application. This class provides
    the basic infrastructure for registering and unregistering clients via a message
    queue.

    Attributes:
        ab_queue (Queue): A multiprocessing queue used for communication between the
        API component and its clients.
        logger (Logger): A logging.Logger instance for logging messages. The logger
        is named after the API.

    Methods:
        __init__(ab_queue: "Queue[Event]"): Constructor for APIBase.
        name: Abstract property. Must be implemented to return the name of the API.
        register(): Attempts to register a client and returns a connection and key.
        type: Abstract property. Must be implemented to return the type of the API.
        unregister(client_pipe: Pipe, key: str): Unregisters a client based on the key.

    Abstract Methods:
        name: Should return the name of the API, mainly used for logging purposes.
        type: Should return the specific ApiType of the API.

    Usage:
        This class must be subclassed and the abstract properties `name` and `type`
        must be implemented. The register and unregister methods provide a mechanism
        for client management, including initiating and closing communication channels.
    """
    def __init__(self, ab_queue: "Queue[Event]") -> None:
        """
        Initializes the APIBase instance with a communication queue.

        Parameters:
            ab_queue (Queue): The multiprocessing queue for inter-process communication.
        """
        super().__init__()
        self.ab_queue = ab_queue
        self.logger = getLogger(self.name)

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Abstract property that should return the name of the API. This name is used
        primarily for logging purposes.

        Returns:
            str: The name of the API.
        """
        pass

    def register(self) -> Tuple[Connection, str] | None:
        """
        Attempts to register a new client with the API. This involves sending a
        registration request through the queue and waiting for a response.

        Returns:
            Tuple[Connection, str] | None: A tuple containing the client connection
            and a registration key if registration is successful; None otherwise.
        """
        client_pipe, client_ab_pipe = Pipe()
        self.ab_queue.put(Register(socket=client_ab_pipe, type=self.type))
        if not client_pipe.poll(timeout=10):
            self.logger.critical("Client registration timeout!")

            client_pipe.close()
            return None
        message = client_pipe.recv()
        if not isinstance(message, Register):
            self.logger.critical("Client registration failed!")
            client_pipe.close()
            return None
        if not message._key:
            self.logger.critical("Key is missing, client registration failed!")
            client_pipe.close()
            return None
        client_ab_pipe.close()
        self.logger.debug("Client was registered %s", message._key)
        return client_pipe, message._key

    @property
    @abstractmethod
    def type(self) -> ApiType:
        """
        Abstract property that should return the type of the API, as defined by the
        ApiType enum. This type is used to identify the API's functionality.

        Returns:
            ApiType: The type of the API.
        """
        pass

    def unregister(self, client_pipe: Connection, key: str) -> None:
        """
        Unregisters a client based on the provided key and closes their communication pipe.

        Parameters:
            client_pipe (Connection): The communication pipe for the client.
            key (str): The registration key of the client to be unregistered.
        """
        unregister = Unregister()
        unregister._key = key
        self.ab_queue.put(unregister)
        client_pipe.close()
        self.logger.debug("Client has left - %s", key)



from logging import getLogger
from pickle import dumps, loads
from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR, SOL_SOCKET, SO_KEEPALIVE, IPPROTO_TCP, TCP_KEEPIDLE, \
    TCP_KEEPINTVL, TCP_KEEPCNT
from time import sleep

from edri.dataclass.event import Event
from edri.config.setting import SWITCH_HOST, SWITCH_PORT

SWITCH_BYTES_LENGTH = 8
SWITCH_MAX_SIZE = 2 ** SWITCH_BYTES_LENGTH


class Socket:

    def __init__(self) -> None:
        self.logger = getLogger('Connector.socket')

    def reconnect(self) -> None:
        while True:
            self.logger.debug("Connecting to switch...")
            self.switch_socket = socket(AF_INET, SOCK_STREAM)
            self.switch_socket.settimeout(10)
            self.switch_socket.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 1)
            self.switch_socket.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
            try:
                self.switch_socket.connect((SWITCH_HOST, SWITCH_PORT))
                self.switch_socket.settimeout(None)
                self.logger.info("Connected")
                break
            except ConnectionRefusedError as e:
                self.logger.warning("Connection failed - connection was refused", exc_info=e)
                sleep(5)
            except TimeoutError as e:
                self.logger.warning("Connection failed - connection takes too long", exc_info=e)
                sleep(5)
            except Exception as e:
                self.logger.warning("Connection failed - unknown error", exc_info=e)
                sleep(10)

    def send(self, event: Event) -> None:
        data = dumps(event)
        message = len(data).to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False)
        message += data
        if len(message) > SWITCH_MAX_SIZE:
            self.logger.error("Message is too big!")
            return
        self.switch_socket.send(message)

    def _load(self, amount: int) -> bytes:
        data = bytes()
        while len(data) < amount:
            data += self.switch_socket.recv(amount - len(data))

        return data

    def receive(self) -> Event:
        while True:
            message_size_bytes = self._load(8)
            if not message_size_bytes:
                self.switch_socket.shutdown(SHUT_RDWR)
                self.switch_socket.close()
                self.logger.error("Received no message size")
                raise ConnectionError

            message_size = int.from_bytes(message_size_bytes, "big", signed=False)
            self.logger.debug("Receiving size message: %s", message_size)
            message = self._load(message_size)
            try:
                event: Event = loads(message)
            except ModuleNotFoundError as e:
                self.logger.error("Event file was not found", exc_info=e)
            except AttributeError as e:
                self.logger.error("Event is not the same", exc_info=e)
            except Exception as e:
                self.logger.critical("Unknown error while loading Event", exc_info=e)
            else:
                self.logger.debug("Received: %s", event)

                event._switch = True
                return event
